/** \file pixel.c
 * \brief    
 * \par
 *
 * \author ARV
 * \n \date	6 . 2017 .
 * \par
 * \copyright Copyright 2015  ARV. All rights reserved.
 * \par
 *
 * \addtogroup PIX
 * @{
 */

#include <avr/io.h>
#include "avr_helper.h"
#include <stdlib.h>
#include "global.h"
#include "pixel.h"

uint8_t rep_lin;


///    
pixel_t	pixels[PIXEL_CNT];

/**
 *       ,   ,  
 * ,      (,  
 *  [R:1,G:1,B:1]   1      ).
 * @param id    ( #PIXEL_CNT)
 * @return 1,   ,  - 0
 */
uint8_t is_light_pix(uint8_t id){
	return !!((uint16_t)(
			(pixels[id].r * (pixels[id].bright+1)) |
			(pixels[id].g * (pixels[id].bright+1)) |
			(pixels[id].b * (pixels[id].bright+1)))>>8
	);
}

/**
 *      
 * @return   
 */
uint8_t get_light_cnt(void){
	uint8_t sum = 0;
	for(uint8_t i=0; i<PIXEL_CNT; i++)
		sum += is_light_pix(i);
	return sum;
}

/**
 *    
 * @param id    ( #PIXEL_CNT)
 * @param bright   
 * @param fade     
 * \note  fade  ,      ,
 *          .
 *    fade  .  fade  0, 
 *       
 */
void bright_ctrl(uint8_t id, uint8_t bright, int8_t fade){
	pixels[id].bright = bright;
	pixels[id].fade = fade;
}

/**
 *     RGB
 * @param pos    ( #PIXEL_CNT)
 * @param r    
 * @param g    
 * @param b    
 */
void set_rgb_color(uint8_t pos, uint8_t r, uint8_t g, uint8_t b){
	pixels[pos].r = r;
	pixels[pos].g = g;
	pixels[pos].b = b;
}

void set_hsv_color(uint8_t pos, hsv_t hsv){
	rgb_t rgb;
	hsv_to_rgb(&hsv, &rgb);
	set_rgb_color(pos, rgb.r, rgb.g, rgb.b);
}

/**
 *            
 * 
 * @param fix_cid    (. #fix_colors)
 * @param fade     
 * \note       , ..   
 *  ,    ,   .
 *         , 
 *     ,  <b>  </b> -
 *    !
 */
void rnd_pix(uint8_t fix_cid, int8_t fade){
	uint8_t id;

	do id = rand() % PIXEL_CNT; while (is_light_pix(id));

	bright_ctrl(id, fade >= 0 ? 255 : 0, fade);
	set_fix_color(id, fix_cid);
}

void rnd_pix_rgb(uint8_t _r, uint8_t _g, uint8_t _b, int8_t fade){
	uint8_t id;

	do id = rand() % PIXEL_CNT; while (is_light_pix(id));

	bright_ctrl(id, fade >= 0 ? 255 : 0, fade);
	set_rgb_color(id, _r, _g, _b);
}

void off_all_pixels(void){
	for(uint8_t i=0; i < PIXEL_CNT; i++)
		bright_ctrl(i, 0, 0);
}
/**
 * @}
 */
