/** \file sd_script.c
 * \brief \b "SCRIPT PLAYER"
 * \n M    SD-
 *
 *  ,   .     ,
 *      .
 * \author ARV
 * \date	15 . 2017 .
 * \copyright Copyright 2015  ARV. All rights reserved.
 *
 *   :\n
 * 	-# AVR GCC 5.2.1    
 *
 * \addtogroup BGEFFECTS
 * \copybrief sd_script.c . \ref sd_script.c
 */

#include <avr/io.h>
#include <avr_helper.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include "global.h"
#include "pixel.h"
#include "color_transform.h"
#include "avr_helper.h"
#include "hardware.h"
#include <util/delay.h>
#include "PFFS/pff.h"
#include "sd_script.h"
#include "script_str.h"
#include "com_io.h"

//  -  WS2812b
#define enter_sd_mode()		do { SPCR = _BV(SPE) | _BV(MSTR); PORT(WS_LOCK_PORT) &= ~WS_LOCK_PIN;} while(0)
//  -  WS2812b
#define leave_sd_mode()		do { SPCR = _BV(SPE) | _BV(MSTR) | _BV(CPHA); SPDR=0;  while(!(SPSR & (1<<SPIF))); PORT(WS_LOCK_PORT) |= WS_LOCK_PIN; } while(0)

///  
typedef struct{
	char		filename[13];	///<   
	uint32_t	pos;			///<      
	uint16_t 	readed;			///<     - 
	char 		*cmd;
} script_t;

///  
static script_t	script;

///  -     .   ,    256
#define SD_BUF_SZ	256UL

///    
typedef enum{
	SCRIPT_OK,			///<  
	SCRIPT_NOTFOUND,	///<           
	SCRIPT_ERROR_IO,	///<  
} script_result_t;

static script_result_t	sdcard_open(void);
static script_result_t	script_open(script_t *s);

///      
static script_result_t script_read_str(script_t *s, char *str);
static uint8_t get_script_digit(char *str);
static char* skip_cmd(char *s);
static char* get_cmd(char *s);

static int8_t select_next = 0;

///  -    ,  
#define VAR_CNT		(LAST_VAR_CHAR - 'A' + 11)
///     
#define STACK_DEPTH	5

typedef struct{
	//    
		uint16_t	total_files;
		uint16_t	file_num;
		uint8_t		random_play;
		uint8_t		pause;
		uint8_t 	ignore;
		uint8_t		inf_break;
} config_t;

static config_t cfg;

///   
typedef struct{
	union{
		pixel_t pixel;
		struct{
			rgb_t		color;			///< 
			uint8_t		bright;			///< 
			int8_t		fade;			///< 
		};
	};
	uint8_t		var[VAR_CNT];	///< 
	uint8_t		rev;			///< 
	uint8_t		stack_head;		///< ""  
	uint32_t	loop_stack[STACK_DEPTH];///<  
} script_param_t;

///  
static script_param_t	param;
///    
static char buf[SD_BUF_SZ];
///  
static script_result_t status;

///  
static FATFS	fs;

uint8_t play_limit;

static uint16_t get_cfg_val(char *src){
	uint16_t result = 0;
	uint8_t tmp;

	while((tmp = get_script_digit(src++)) < 10)
		result = result*10 + tmp;

	return result;
}

#if defined(_DEBUG_)
static void debug_out(uint16_t d){
	uint16_t mask=0x8000;
	  for(uint8_t i=0; i<16; i++, mask>>=1){
		  putch(d & mask ? '1' : '0');
	  }
	  putch('\r'); putch('\n');
}

static void debug_str(const __flash char *str){
	while(*str) putch(*str++);
    //putch('\r'); putch('\n');
}

void debug_cmd(char *cmd){
	while(*cmd > ' ') putch(*cmd++);
    putch('\r'); putch('\n');
}

static void debug_num(const __flash char *str, uint16_t num){
	debug_str(str); putch(':');
	char tmp[7];

	itoa(num, tmp, 10);
	debug_cmd(tmp);
}
#else
#define debug_num(a,b)
#define debug_cmd(o)
#define debug_str(i)
#define debug_out(u)
#endif



#define cmd_is(s) (strncmp_P(cmd, PSTR(s), sizeof(s)-1) == 0)

static void get_config(void){
	uint16_t readed;
	char *cmd;

	cfg.total_files = 0;
	cfg.file_num = -1; //  -1, .. ,       -   .
	cfg.inf_break = 0;
	cfg.random_play = 0;
	cfg.pause = 0;
	cfg.ignore = 0;
	rep_lin = 1;

	if(pf_open(SCR_CFG_FILE) == FR_OK){
		//  
		debug_led_on();
		if(pf_read(&buf, 255, &readed) == FR_OK){
			cmd = get_cmd(buf);
			while((cmd - buf) < readed){
				if(cmd == NULL) break;
				if(cmd[2] != '=') break;
				debug_cmd(cmd);
				if(cmd_is(SCR_CFG_TF)){
					cfg.total_files = get_cfg_val(cmd+3);
				} else if(cmd_is(SCR_CFG_IG)){
					cfg.ignore = get_script_digit(cmd+3);
					if(cfg.ignore > 1) cfg.ignore = 0;

				}else if(cmd_is(SCR_CFG_RP)){
					cfg.random_play = get_script_digit(cmd+3);
					if(cfg.random_play > 1) cfg.random_play = 0;

				} else if(cmd_is(SCR_CFG_IB)){
					cfg.inf_break = get_cfg_val(cmd+3);

				} else if(cmd_is(SCR_CFG_PS)){
					cfg.pause = get_script_digit(cmd+3);
					if(cfg.pause > 1) cfg.pause = 0;

				} else if(cmd_is(SCR_CFG_LR)){
					rep_lin = get_script_digit(cmd+3);
					if(rep_lin > 4) rep_lin = 1;
				}
				cmd = get_cmd(skip_cmd(cmd));
			}
		}
	}
	debug_led_off();
}

/**  SD-
 *
 * @return  
 */
static script_result_t	sdcard_open(void){
	//   
	if(pf_mount(&fs) == FR_OK){
		//debug_str(PSTR("CFG:"));
		get_config();
		return cfg.total_files == 0 ? SCRIPT_NOTFOUND : SCRIPT_OK;
	}
	//     
	return SCRIPT_NOTFOUND;
}


/**
 *
 * @param s    
 * @return  
 */
static script_result_t	script_open(script_t *s){
	script_result_t result;

	memset(s, 0, sizeof(script_t));
	memset(&param, 0, sizeof(script_param_t));
	memset(buf, 0, SD_BUF_SZ);

	do{
		wdt_reset();
		//   
		if(cfg.random_play){
			cfg.file_num = rand() % cfg.total_files;
		}else{
			cfg.file_num++;
			//     
			if (cfg.file_num >= cfg.total_files){
				result = SCRIPT_NOTFOUND;
				break;
			}
		}
		itoa(cfg.file_num, s->filename, 10);
		strcat_P(s->filename,PSTR(SCR_FILE_EXT));
		//    
		debug_cmd(s->filename);

		debug_led_on();
		result = pf_open(s->filename)== FR_OK ? SCRIPT_OK : SCRIPT_NOTFOUND;
		debug_led_off();
		//    ,     
	} while (!cfg.random_play && cfg.ignore && (result == SCRIPT_NOTFOUND));

	if(result == SCRIPT_OK){
		play_limit = cfg.inf_break;
	}

	return result;
}

/**
 *     ,       0
 * @param s   
 * @param str   
 * @return  
 */
static script_result_t script_read_str(script_t *s, char *str){
	uint16_t n = SD_BUF_SZ-1;
	script_result_t res = SCRIPT_ERROR_IO;

	debug_led_on();
	*str = 0;
	//  
	if(pf_lseek(s->pos) == FR_OK){
		//   
		if(pf_read(str, n, &s->readed) == FR_OK){
			//   - 
			if(s->readed < n) n = s->readed;
			//   ""        

			//  ""  :   -    
			while(n && str[n] && (str[n] != ' ') && (str[n] != '\n') && (str[n] != '\r') && (str[n] != '\t')) {
				n--;
			}
			//   
			str[n] = 0;
			//      
			s->readed = n;
			s->pos += n;
			res = SCRIPT_OK;
		}
	}
	debug_led_off();
	return res;
}

/**
 *   
 * @param str     
 * @return  0...15  255,   
 */
static uint8_t get_script_digit(char *str){
	if((str[0] >= '0') && (str[0] <= '9')) return str[0] - '0';
	else if((str[0] >= 'A') && (str[0] <= LAST_VAR_CHAR)) return str[0] - 'A' + 10;
	else return 255;
}

#define ERROR_CH	0xFF00

/**
 *     
 * \note           .
 * @param str   
 * @return     0...255  0xFF00      
 */
static uint16_t get_script_val(char *str){
	//  ?
	if(str[0] == SCR_VAR_CHAR){
		//   
		uint8_t n = get_script_digit(++str);
		//     
		return n < VAR_CNT ? param.var[n] : ERROR_CH;
	} else if (str[0] == SCR_FUNC_RD[0]){
		//  RD  RP
		if(str[1] == SCR_FUNC_RD[1]) return rand() & 0x00FF;
		else if(str[1] == SCR_FUNC_RP[1]) return rand() % PIXEL_CNT;
		else return ERROR_CH;
	} else if (str[0] == SCR_FUNC_TP[0]){
		//  TP
		if(str[1] == SCR_FUNC_TP[1]) return PIXEL_CNT-1;
		else return ERROR_CH;
	} else if(str[0] == SCR_FUNC_DP[0] && str[1] == SCR_FUNC_DP[1]){
		//  DP
		uint8_t px, cnt;
		cnt = PIXEL_CNT; //        - 
		do{
			//   
			px = rand() % PIXEL_CNT;
			//            
		} while((
				((pixels[px].r * pixels[px].bright) >> 8) |
				((pixels[px].g * pixels[px].bright) >> 8) |
				((pixels[px].b * pixels[px].bright) >> 8)) && cnt--);

		return px;
	} else {
		// 
		uint8_t nh = get_script_digit(str++);
		uint8_t nl = get_script_digit(str);
		if((nh < 0x10) && (nl < 0x10)) return (nh << 4) | nl;
		else return ERROR_CH;
	}
}

/**
 *      
 * @return  
 */
static uint32_t popstack(void){
	//    
	if(param.stack_head != 0){
		//  ""
		param.stack_head--;
		//  
		return param.loop_stack[param.stack_head];
	}
	return 0;
}

/**
 *    
 * @param d    
 */
static void pushstack(uint32_t d){
	//     
	if(param.stack_head < STACK_DEPTH){
		param.loop_stack[param.stack_head++] = d;
	}
	//       -   ,    
}

/**
 *     ,     .
 */
static void do_cache_jump(void){
	uint32_t to;
	to = popstack();
	if(to >= (script.pos - script.readed)){
		script.cmd = buf + (to - script.pos + script.readed);
	} else {
		//    seek,    
		script.pos = to;
		//       
		script.cmd[0] = 0;
		//debug_num(PSTR("pos"),script.pos);
		//debug_num(PSTR("seek to"),to);
	}
	pushstack(to);
}

/**
 *    
 * @param cmd     
 * @return  ,       
 */
static uint8_t execute_cmd(char *cmd){
	uint16_t tmp;

	//      
	do{
		//     
		if(cmd[0] == 'V'){
			//  
			uint8_t n = get_script_digit(cmd+1);
			if(n >= VAR_CNT) break;
			//  
			tmp = get_script_val(cmd+3);
			if(tmp > 255) break;
			//   
			switch(cmd[2]){
			case '=': param.var[n] = tmp; break;
			case '+': tmp += param.var[n]; param.var[n] = tmp > 255 ? 255 : tmp; break;
			case '-': tmp = param.var[n] - tmp; param.var[n] = tmp < 0 ? 255: tmp; break;
			case '*': tmp *= param.var[n]; param.var[n] = tmp > 255 ? 255 : tmp; break;
			case '/': if(tmp == 0) return 0;
					  param.var[n] /= tmp;
					  break;
			case '%': if(tmp == 0) return 0;
					  param.var[n] %= tmp;
					  break;
		    //    
			//case '>' param.var[n] = param.var[n] > tmp; break;
			//case '<' param.var[n] = param.var[n] < tmp; break;
			//case '=' param.var[n] = param.var[n] == tmp; break;
			//case '!' param.var[n] = param.var[n] != tmp; break;
			default:
				//  
				break;
			}
			//  ,   V,      - 
			break;
		} else
		//     
		if(cmd[2] == '='){
			//  
			tmp = get_script_val(cmd+3);
			if(tmp > 255) break;
			//    -  
			if(cmd_is(SCR_CMD_PB)){
				// 
				param.bright = tmp;

			} else if(cmd_is(SCR_CMD_GB)){
				//  
				for(uint8_t i=0; i<PIXEL_CNT; i++)
					pixels[i].bright = tmp;

			} else if(cmd_is(SCR_CMD_PF)){
				// 
				param.fade = tmp;

			} else if(cmd_is(SCR_CMD_GF)){
				//  
				for(uint8_t i=0; i<PIXEL_CNT; i++)
					pixels[i].fade = tmp;

			} else if(cmd_is(SCR_CMD_PC)){
				// 
				uint16_t tmp2 = get_script_val(cmd+5);
				uint16_t tmp3 = get_script_val(cmd+7);
				if((tmp2 > 255) || (tmp3 > 255)) break;
				param.color.r = tmp;
				param.color.g = tmp2;
				param.color.b = tmp3;

			} else if(cmd_is(SCR_CMD_GC)){
				//  
				uint16_t tmp2 = get_script_val(cmd+5);
				uint16_t tmp3 = get_script_val(cmd+7);
				if((tmp2 > 255) || (tmp3 > 255)) break;
				for(uint8_t i=0; i<PIXEL_CNT; i++)
					set_rgb_color(i, tmp, tmp2, tmp3);

			} else if(cmd_is(SCR_CMD_GI)){
				//    
				rgb_t rgb;
				rgb_from_hue(tmp, &rgb);
				for(uint8_t i=0; i<PIXEL_CNT; i++)
					pixels[i].rgb = rgb;
			} else if(cmd_is(SCR_CMD_PI)){
				//    
				rgb_t rgb;
				rgb_from_hue(tmp, &rgb);
				param.color = rgb;
			} else if(cmd_is(SCR_CMD_PM)){
				//  
				char *ptr = cmd+5;
				while(tmp <= 255){
					if(tmp < PIXEL_CNT){
						if(param.rev) tmp = PIXEL_CNT - tmp - 1;
						//     
						pixels[tmp] = param.pixel;
					} //       
					tmp = get_script_val(ptr);
					ptr += 2;
				}

			} else if(cmd_is(SCR_CMD_BM)){
				//  
				char *ptr = cmd+5;
				while(tmp <= 255){
					if(tmp < PIXEL_CNT){
						if(param.rev) tmp = PIXEL_CNT - tmp - 1;
						//     
						pixels[tmp].bright = param.bright;
					} //       
					tmp = get_script_val(ptr);
					ptr += 2;
				}

			} else if(cmd_is(SCR_CMD_FM)){
				//  
				char *ptr = cmd+5;
				while(tmp <= 255){
					if(tmp < PIXEL_CNT){
						if(param.rev) tmp = PIXEL_CNT - tmp - 1;
						//     
						pixels[tmp].fade = param.fade;
					} //       
					tmp = get_script_val(ptr);
					ptr += 2;
				}

			} else if(cmd_is(SCR_CMD_WT)){
				// 
				return tmp;
			}
		} else
		//     
		if(cmd_is(SCR_CMD_LV)){
			//   
			uint8_t n = get_script_digit(cmd+2);
			if(n > VAR_CNT) break;
			//  
			uint8_t to_do = 0; //   
			uint8_t ch3 = cmd[3];

			if((ch3 == '=') || (ch3 == '>') || (ch3 == '<') || (ch3 == '!')){
				tmp = get_script_val(cmd+4);
				if(tmp > 255) break;
				switch(ch3){
				case '=': to_do = (param.var[n] == tmp);  break;
				case '>': to_do = (param.var[n] > tmp);	  break;
				case '<': to_do = (param.var[n] < tmp);	  break;
				case '!': to_do = (param.var[n] != tmp);  break;
				}
			} else {
				//   ,     
				if(param.var[n] != 0){
					param.var[n]--;
					to_do = 1;
				}
			}
			//     
			if(to_do)	do_cache_jump();	//     
			else		popstack();			//     
			break;
		} else
		//     
		if(cmd_is(SCR_CMD_END)){
			//  
			cmd[0] = 0;
			status = SCRIPT_ERROR_IO;

		} else if(cmd_is(SCR_CMD_CLR)){
			// 
			memset(pixels, 0, PIXEL_CNT * sizeof(pixel_t));
			memset(&param, 0, sizeof(script_param_t));
			param.bright = 0xFF;

		} else if(cmd_is(SCR_CMD_INF)){
			//  
			do_cache_jump();

		} else if(cmd_is(SCR_CMD_RST)){
			// 
			script.pos = 0;
			cmd[0] = 0; // !!!! !!!

		} else if(cmd_is(SCR_CMD_RPT)){
			//  
			//   !!!!
			uint32_t next = script.pos - (script.readed - (cmd - buf + 3));
			//debug_num(PSTR("rep-start"),next);
			pushstack(next);

		} else if(cmd_is(SCR_CMD_PNT)){
			// 
			return 1;
		}else if(cmd_is(SCR_CMD_REV)){
			// 
			param.rev = !param.rev;

		} else if(cmd_is(SCR_CMD_RGC)){
			//   
			rgb_t rgb;
			rgb_from_hue(rand() % HSV_GRADE, &rgb);
			for(uint8_t i=0; i<PIXEL_CNT; i++)
				pixels[i].rgb = rgb;

		} else if(cmd_is(SCR_CMD_RPC)){
			//   
			rgb_t rgb;
			rgb_from_hue(rand() % HSV_GRADE, &rgb);
			param.color = rgb;

		} else if(cmd_is(SCR_CMD_NEG)){
			//  
			for(uint8_t i=0; i<PIXEL_CNT; i++)
				pixels[i].bright ^= 0xFF;

#if defined(INCLUDE_SHIFT_CMD)
		} else if(cmd_is(SCR_CMD_RRC)){
			//     
			rgb_t c = pixels[PIXEL_CNT-1].rgb;
			for(uint8_t i=PIXEL_CNT-1; i>=1; i--)
				pixels[i].rgb = pixels[i-1].rgb;
			pixels[0].rgb = c;

		} else if(cmd_is(SCR_CMD_RLC)){
			//     
			rgb_t c = pixels[0].rgb;
			for(uint8_t i=0; i<(PIXEL_CNT-1); i++)
				pixels[i].rgb = pixels[i+1].rgb;
			pixels[PIXEL_CNT-1].rgb = c;

		} else if(cmd_is(SCR_CMD_RRB)){
			//     
			uint8_t b = pixels[PIXEL_CNT-1].bright;
			for(uint8_t i=PIXEL_CNT-1; i>=1; i--)
				pixels[i].bright = pixels[i-1].bright;
			pixels[0].bright = b;

		} else if(cmd_is(SCR_CMD_RLB)){
			//     
			uint8_t b = pixels[0].bright;
			for(uint8_t i=0; i<(PIXEL_CNT-1); i++)
				pixels[i].bright = pixels[i+1].bright;
			pixels[PIXEL_CNT-1].bright = b;

		} else if(cmd_is(SCR_CMD_SRC)){
			//   
			for(uint8_t i=PIXEL_CNT-1; i>=1; i--)
				pixels[i].rgb = pixels[i-1].rgb;
			pixels[0].r = 0;
			pixels[0].g = 0;
			pixels[0].b = 0;

		} else if(cmd_is(SCR_CMD_SLC)){
			//   
			for(uint8_t i=0; i<(PIXEL_CNT-1); i++)
				pixels[i].rgb = pixels[i+1].rgb;
			pixels[PIXEL_CNT-1].r = 0;
			pixels[PIXEL_CNT-1].g = 0;
			pixels[PIXEL_CNT-1].b = 0;

		} else if(cmd_is(SCR_CMD_SRB)){
			//   
			for(uint8_t i=PIXEL_CNT-1; i>=1; i--)
				pixels[i].bright = pixels[i-1].bright;
			pixels[0].bright = 0;

		} else if(cmd_is(SCR_CMD_SLB)){
			//   
			for(uint8_t i=0; i<(PIXEL_CNT-1); i++)
				pixels[i].bright = pixels[i+1].bright;
			pixels[PIXEL_CNT-1].bright = 0;
#endif
		} else {
			//    
			//debug_cmd(cmd);
		}
	} while(0);

	return 0;
}

/**
 *  
 */
static void start(void){
	//  
	off_all_pixels();
	//   
	status = sdcard_open();
	if(status == SCRIPT_OK){
		//    
		status = script_open(&script);
	}
}

/**
 *    
 * @param s     
 * @return      -
 */
static char* skip_cmd(char *s){
	uint8_t n = 0;
	//     
	while(s[n] && (s[n] != ' ') && (s[n] != '\n') && (s[n] != '\r') && (s[n] != '\t')) n++;
	return s + n;
}

/**
 *      
 * @param s   
 * @return        NULL,   
 * 
 */
static char* get_cmd(char *s){
	uint8_t n = 0;
	//    
	while(s[n] && ((s[n] == ' ') || (s[n] == '\n') || (s[n] == '\r') || (s[n] == '\t'))) n++;
	return s[n] ? s + n : NULL;
}

/**
 *       
 * @param buf   
 * @return  ,     
 */
uint8_t script_execute(char *buf){
	uint8_t result;

	do{
		if(!play_limit && cfg.inf_break){
			//      , 
			//    
			status = SCRIPT_ERROR_IO;
			break;
		}
		if(script.cmd == NULL){
			//    ,   
			debug_led_on();
			status = script_read_str(&script, buf);
			debug_led_off();
			//     
			script.cmd = get_cmd(buf);
			continue;
		}
		//   
		result = execute_cmd(script.cmd);
		//   
		script.cmd = get_cmd(skip_cmd(script.cmd));
		//   -    
		if(result) return result;
	}while(status == SCRIPT_OK);

	return 0;
}

/**
 *         
 * @param s      ()
 */
void play_script(void){
	static uint8_t div;

	wdt_reset();

	if(div){
		//    
		--div;
		return;
	}

	enter_sd_mode();

	if(select_next != 0){
		goto next_file_load;
	}

	switch(status){
	case SCRIPT_OK:
		//   
		div = script_execute(buf);
		break;
	case SCRIPT_NOTFOUND:
		start();
		break;

	case SCRIPT_ERROR_IO:
		//     -     
		//    ,  -  
		if(cfg.pause){
			//    
			div = 150;
			for(uint8_t i=0; i<PIXEL_CNT; i++)
				pixels[i].fade = 10;
		}
		if(select_next == 0){
			select_next = 1;
		}
		break; // !!!!! ,          
	next_file_load:
		//    
		//  
		status = script_open(&script);
		select_next = 0;

	default:
		break;
	}

	leave_sd_mode();
}

INIT(7){
	leave_sd_mode();
}

