/** \file ws2812.c
 * \brief   WS2812    SPI
 *
 * \author ARV
 * \date	2 . 2017 .
 * \copyright Copyright 2015  ARV. All rights reserved.
 * \par
 * \defgroup WS  WS2812    SPI
 * \brief     WS2812
 *
 * \addtogroup WS
 * @{
 */

#include <avr/io.h>
#include "avr_helper.h"
#include <util/delay.h>
#include "global.h"
#include "pixel.h"
#include "ws2812.h"

#if (F_CPU != 16000000UL) || !defined(F_CPU)
#error "F_CPU must be defined and must be 16MHz"
#endif

///     "0"  WS2812
#define WS_BIT_0	0b11000000
///     "1"  WS2812
#define WS_BIT_1	0b11111100

/**   1    WS2812
 *
 * @param byte  
 */
__inline static void send_byte(uint8_t byte){
	for(uint8_t mask=0x80; mask; mask >>=1){
		while(bit_is_clear(SPSR, SPIF));
		SPDR = byte & mask ? WS_BIT_1 : WS_BIT_0;
	}
}

/**     WS2812
 *
 */
void ws2812_show(void){
	uint16_t b;
	for(uint8_t i=rep_lin; i; i--){
		for(uint8_t i=0; i<PIXEL_CNT; i++){
			b = pixels[i].bright+1;
			//   WS2812      :
			send_byte((pixels[i].g * b)>>8);	//  
			send_byte((pixels[i].r * b)>>8);	//  
			send_byte((pixels[i].b * b)>>8);	//   - 
		}
	}
}

/**   SPI
 *
 */
INIT(7){
	SPCR = _BV(SPE) | _BV(MSTR) | _BV(CPHA);//fosc/2 16MHz/2
	SPSR = _BV(SPI2X);
	SPSR; SPDR;
}

/**
 * @}
 */
